import asyncio
from contextlib import contextmanager
from typing import Any, Callable
from unittest import mock


async def run(func: Callable, *args: Any) -> Any:
    return await asyncio.get_event_loop().run_in_executor(None, func, *args)


async def _in_thread_execution(func: Callable, *args: Any) -> Any:
    return func(*args)


@contextmanager
def mock_executor():
    """Mock executor so we can include in coverage which has concurrency
    set to greenlet
    """
    with mock.patch(
        "fluid.executor.run",
        _in_thread_execution,
    ):
        yield



from __future__ import annotations

import asyncio
import logging
from collections import defaultdict, deque
from contextlib import AsyncExitStack
from functools import partial
from typing import Any, Awaitable, Callable, Self

from inflection import underscore
from typing_extensions import Annotated, Doc

from fluid.utils import log
from fluid.utils.dispatcher import AsyncDispatcher, Dispatcher, Event
from fluid.utils.worker import AsyncConsumer, WorkerFunction, Workers

from .broker import TaskBroker, TaskRegistry
from .errors import TaskAbortedError, TaskRunError, UnknownTaskError
from .models import (
    Task,
    TaskManagerConfig,
    TaskPriority,
    TaskRun,
    TaskRunWaiter,
    TaskState,
)

AsyncHandler = Callable[[TaskRun], Awaitable[None]]

logger = log.get_logger(__name__)


class TaskDispatcher(Dispatcher[TaskRun]):
    """The task dispatcher is responsible for dispatching task run messages"""

    def message_type(self, message: TaskRun) -> str:
        return message.state


class AsyncTaskDispatcher(AsyncDispatcher[TaskRun]):

    def message_type(self, message: TaskRun) -> str:
        return message.state


class TaskManager:
    """The task manager is the main entry point for managing tasks"""

    def __init__(self, **kwargs: Any) -> None:
        self.state: dict[str, Any] = {}
        self.config: TaskManagerConfig = TaskManagerConfig(**kwargs)
        self.dispatcher: Annotated[
            TaskDispatcher,
            Doc(
                """
                A dispatcher of task run events.

                Register handlers to listen for task run events.
                """
            ),
        ] = TaskDispatcher()
        self.broker = TaskBroker.from_url(self.config.broker_url)
        self._stack = AsyncExitStack()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, *exc_info: Any) -> None:
        try:
            await self._stack.aclose()
        finally:
            await self.on_shutdown()

    async def enter_async_context(self, cm: Any) -> Any:
        return await self._stack.enter_async_context(cm)

    @property
    def registry(self) -> TaskRegistry:
        return self.broker.registry

    @property
    def type(self) -> str:
        return underscore(self.__class__.__name__)

    async def execute(self, task: Task | str, **params: Any) -> TaskRun:
        """Execute a task and wait for it to finish"""
        task_run = self.create_task_run(task, **params)
        await task_run.execute()
        return task_run

    async def on_shutdown(self) -> None:
        await self.broker.close()

    def execute_sync(self, task: Task | str, **params: Any) -> TaskRun:
        return asyncio.run(self._execute_and_exit(task, **params))

    def register_task(self, task: Task) -> None:
        """Register a task with the task manager

        Only tasks registered can be executed by a task manager
        """
        self.broker.register_task(task)

    async def queue(
        self,
        task: str | Task,
        priority: TaskPriority | None = None,
        **params: Any,
    ) -> TaskRun:
        """Queue a task for execution

        This methods fires two events:

        - queue: when the task is about to be queued
        - queued: after the task is queued
        """
        task_run = self.create_task_run(task, priority=priority, **params)
        self.dispatcher.dispatch(task_run)
        task_run.set_state(TaskState.queued)
        await self.broker.queue_task(task_run)
        return task_run

    def create_task_run(
        self,
        task: str | Task,
        run_id: str = "",
        priority: TaskPriority | None = None,
        **params: Any,
    ) -> TaskRun:
        """Create a TaskRun in `init` state"""
        if isinstance(task, str):
            task = self.broker.task_from_registry(task)
        run_id = run_id or self.broker.new_uuid()
        return TaskRun(
            id=run_id,
            task=task,
            priority=priority or task.priority,
            params=params,
            task_manager=self,
        )

    def register_from_module(self, module: Any) -> None:
        for name in dir(module):
            if name.startswith("_"):
                continue
            if isinstance(obj := getattr(module, name), Task):
                self.register_task(obj)

    def register_from_dict(self, data: dict) -> None:
        for name, obj in data.items():
            if name.startswith("_"):
                continue
            if isinstance(obj, Task):
                self.register_task(obj)

    def register_async_handler(self, event: str, handler: AsyncHandler) -> None:
        """Register an async handler for a given event

        This method is a no op for a TaskManager that is not a worker
        """

    def unregister_async_handler(self, event: Event | str) -> AsyncHandler | None:
        """Unregister an async handler for a given event

        This method is a no op for a TaskManager that is not a worker
        """
        return None

    async def _execute_and_exit(self, task: Task | str, **params: Any) -> TaskRun:
        async with self:
            return await self.execute(task, **params)


class TaskConsumer(TaskManager, Workers):
    """The Task Consumer is a Task Manager responsible for consuming tasks
    from a task queue
    """

    def __init__(self, **config: Any) -> None:
        super().__init__(**config)
        Workers.__init__(self)
        self._async_dispatcher_worker = AsyncConsumer(AsyncTaskDispatcher())
        self._concurrent_tasks: dict[str, dict[str, TaskRun]] = defaultdict(dict)
        self._task_to_queue: deque[str | Task] = deque()
        self._priority_task_run_queue: deque[TaskRun] = deque()
        self._queue_tasks_worker = WorkerFunction(
            self._queue_task, name="queue-task-worker"
        )
        self.add_workers(self._queue_tasks_worker)
        self.add_workers(self._async_dispatcher_worker)
        for i in range(self.config.max_concurrent_tasks):
            worker_name = f"task-worker-{i+1}"
            self.add_workers(
                WorkerFunction(
                    partial(self._consume_tasks, worker_name), name=worker_name
                )
            )

    @property
    def num_concurrent_tasks(self) -> int:
        """The number of concurrent_tasks running in the consumer"""
        return sum(len(v) for v in self._concurrent_tasks.values())

    def sync_queue(self, task: str | Task) -> None:
        self._task_to_queue.appendleft(task)

    def sync_priority_queue(self, task: str | Task) -> None:
        self._priority_task_run_queue.appendleft(self.create_task_run(task))

    def num_concurrent_tasks_for(self, task_name: str) -> int:
        """The number of concurrent tasks for a given task_name"""
        return len(self._concurrent_tasks[task_name])

    async def queue_and_wait(
        self, task: str, *, timeout: int = 2, **params: Any
    ) -> TaskRun:
        """Queue a task and wait for it to finish"""
        with TaskRunWaiter(self) as waiter:
            return await waiter.wait(await self.queue(task, **params), timeout=timeout)

    def register_async_handler(self, event: Event | str, handler: AsyncHandler) -> None:
        event = Event.from_string_or_event(event)
        self.dispatcher.register_handler(
            f"{event.type}.async_dispatch",
            self._async_dispatcher_worker.send,
        )
        self._async_dispatcher_worker.dispatcher.register_handler(event, handler)

    def unregister_async_handler(self, event: Event | str) -> AsyncHandler | None:
        return self._async_dispatcher_worker.dispatcher.unregister_handler(event)

    # Internals

    # process tasks from the internal queue
    async def _queue_task(self) -> None:
        try:
            task = self._task_to_queue.pop()
        except IndexError:
            await asyncio.sleep(0.1)
        else:
            await self.queue(task)
            await asyncio.sleep(0)

    async def _consume_tasks(self, worker_name: str) -> None:
        if not self.config.consume_tasks:
            await asyncio.sleep(self.config.sleep)
            return
        if self._priority_task_run_queue:
            task_run = self._priority_task_run_queue.pop()
        else:
            try:
                maybe_task_run = await self.broker.get_task_run(self)
            except UnknownTaskError as exc:
                logger.error(
                    "%s unknown task %s - it looks like it is not "
                    "registered with this consumer",
                    worker_name,
                    exc,
                )
                maybe_task_run = None
            if not maybe_task_run:
                return
            else:
                task_run = maybe_task_run
        task_name = task_run.name
        self._concurrent_tasks[task_name][task_run.id] = task_run
        #
        if (
            task_run.task.max_concurrency > 0
            and task_run.task.max_concurrency < self.num_concurrent_tasks_for(task_name)
        ):
            task_run.set_state(TaskState.rate_limited)
        elif not (await self.broker.get_tasks_info(task_name))[0].enabled:
            task_run.set_state(TaskState.aborted)
        #
        else:
            try:
                params = task_run.params_dump_json()
            except Exception:
                task_run.logger.exception("%s - start - params exeception", task_run.id)
            else:
                task_run.logger.info("%s - %s - start", task_run.id, params)
            try:
                async with asyncio.timeout(task_run.task.timeout_seconds):
                    await task_run.execute()
            except TaskRunError:
                # no logging as this was a controlled exception
                pass
            except TaskAbortedError as exc:
                task_run.logger.info("%s - %s - aborted - %s", task_run.id, params, exc)
            except asyncio.TimeoutError:
                task_run.logger.error("task run %s - %s - timeout", task_run.id, params)
            except Exception:
                task_run.logger.exception("critical exception while executing")

        self._concurrent_tasks[task_name].pop(task_run.id, None)
        duration_ms = task_run.duration_ms
        if duration_ms is not None:
            await self.broker.update_task(
                task_run.task,
                dict(
                    last_run_end=task_run.end,
                    last_run_duration=task_run.duration,
                    last_run_state=task_run.state,
                ),
            )
            task_run.logger.log(
                logging.WARNING if task_run.is_failure else logging.INFO,
                "end - %s - milliseconds - %s",
                task_run.state,
                duration_ms,
            )


# required by pydantic to avoid `Class not fully defined` error
TaskRun.model_rebuild()



import logging
from logging.config import dictConfig
from typing import Any, Sequence

try:
    import pythonjsonlogger
except ImportError:
    pythonjsonlogger = None  # type: ignore

from fluid import settings

logger = logging.getLogger()


logger = logging.getLogger(settings.APP_NAME)


def get_logger(name: str = "", prefix: bool = False) -> logging.Logger:
    if prefix:
        return logger.getChild(name) if name else logger
    else:
        return logging.getLogger(name) if name else logger


def level_num(level: str) -> int:
    return getattr(logging, level.upper())


def log_config(
    level: int,
    other_level: int = logging.WARNING,
    app_names: Sequence[str] = (settings.APP_NAME,),
    log_handler: str = settings.LOG_HANDLER,
    log_format: str = settings.PYTHON_LOG_FORMAT,
    formatters: dict[str, dict[str, str]] | None = None,
) -> dict:
    other_level = max(level, other_level)
    log_handlers = {
        "plain": {
            "level": level,
            "class": "logging.StreamHandler",
            "formatter": "plain",
        }
    }
    log_formatters: dict = {"plain": {"format": log_format}}
    if pythonjsonlogger is not None:
        log_handlers.update(
            json={
                "level": level,
                "class": "logging.StreamHandler",
                "formatter": "json",
            },
            nicejson={
                "level": level,
                "class": "logging.StreamHandler",
                "formatter": "nicejson",
            },
        )
        log_formatters.update(
            json={
                "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
                "format": log_format,
            },
            nicejson={
                "()": "pythonjsonlogger.jsonlogger.JsonFormatter",
                "format": log_format,
                "json_indent": 2,
            },
        )
    if formatters:
        log_formatters.update(formatters)
    app_name_set = set(app_names)
    if settings.APP_NAME not in app_name_set:
        app_name_set.add(settings.APP_NAME)
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": log_formatters,
        "handlers": log_handlers,
        "loggers": {
            app_name: {"level": level, "handlers": [log_handler], "propagate": 0}
            for app_name in app_name_set
        },
        "root": {"level": other_level, "handlers": [log_handler]},
    }


def config(**kwargs: Any) -> dict:
    cfg = log_config(level_num(settings.LOG_LEVEL), **kwargs)
    dictConfig(cfg)
    return cfg



from datetime import date, datetime
from typing import Any, Set, TypeAlias, cast

from dateutil.parser import parse as parse_date
from sqlalchemy import Column, Table, func, insert, select
from sqlalchemy.engine.cursor import CursorResult
from sqlalchemy.engine.row import Row
from sqlalchemy.ext.asyncio import AsyncConnection
from sqlalchemy.sql import FromClause, Select, and_, or_
from sqlalchemy.sql.dml import Delete, Insert, Update

from fluid.utils.errors import ValidationError

from .container import Database

QueryType: TypeAlias = Delete | Select | Update


class CrudDB(Database):
    """A :class:`.Database` with additional methods for CRUD operations"""

    async def db_select(
        self,
        table: FromClause,
        filters: dict,
        *,
        order_by: tuple[str, ...] | None = None,
        conn: AsyncConnection | None = None,
    ) -> CursorResult:
        """Select rows from a given table
        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        sql_query = self.get_query(table, Select(table), params=filters)
        if order_by:
            sql_query = self.order_by_query(table, cast(Select, sql_query), order_by)
        async with self.ensure_transaction(conn) as conn:
            return await conn.execute(sql_query)

    async def db_insert(
        self,
        table: Table,
        data: list[dict] | dict,
        *,
        conn: AsyncConnection | None = None,
    ) -> CursorResult:
        """Perform an insert into a table
        :param table: sqlalchemy Table
        :param data: key-value pairs for columns values
        :param conn: optional db connection
        """
        async with self.ensure_transaction(conn) as conn:
            sql_query = self.insert_query(table, data)
            return await conn.execute(sql_query)

    async def db_update(
        self,
        table: Table,
        filters: dict,
        data: dict,
        *,
        conn: AsyncConnection | None = None,
    ) -> CursorResult:
        """Perform an update of rows

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows to update
        :param data: key-value pairs for updating columns values of selected rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        update = (
            cast(
                Update,
                self.get_query(table, table.update(), params=filters),
            )
            .values(**data)
            .returning(*table.columns)
        )
        async with self.ensure_transaction(conn) as conn:
            return await conn.execute(update)

    async def db_upsert(
        self,
        table: Table,
        filters: dict,
        data: dict | None = None,
        *,
        conn: AsyncConnection | None = None,
    ) -> Row:
        """Perform an upsert for a single record

        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows to update
        :param data: key-value pairs for updating columns values of selected rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        if data:
            result = await self.db_update(table, filters, data, conn=conn)
        else:
            result = await self.db_select(table, filters, conn=conn)
        record = result.one_or_none()
        if record is None:
            insert_data = data.copy() if data else {}
            insert_data.update(filters)
            result = await self.db_insert(table, insert_data, conn=conn)
            record = result.one()
        return record

    async def db_delete(
        self,
        table: Table,
        filters: dict,
        *,
        conn: AsyncConnection | None = None,
    ) -> CursorResult:
        """Delete rows from a given table
        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        sql_query = self.get_query(
            table,
            table.delete().returning(*table.columns),
            params=filters,
        )
        async with self.ensure_transaction(conn) as conn:
            return await conn.execute(sql_query)

    async def db_count(
        self,
        table: FromClause,
        filters: dict,
        *,
        conn: AsyncConnection | None = None,
    ) -> int:
        """Count rows in a table
        :param table: sqlalchemy Table
        :param filters: key-value pairs for filtering rows
        :param conn: optional db connection
        :param consumer: optional consumer (see :meth:`.get_query`)
        """
        count_query = self.db_count_query(
            cast(
                Select,
                self.get_query(
                    table,
                    table.select(),
                    params=filters,
                ),
            ),
        )
        async with self.ensure_connection(conn) as conn:
            result: CursorResult = await conn.execute(count_query)
            return cast(int, result.scalar())

    # Query methods

    def insert_query(self, table: Table, records: list[dict] | dict) -> Insert:
        if isinstance(records, dict):
            records = [records]
        else:
            cols: Set[str] = set()
            for record in records:
                cols.update(record)
            new_records = []
            for record in records:
                if len(record) < len(cols):
                    record = record.copy()
                    missing = cols.difference(record)
                    for col in missing:
                        record[col] = None
                new_records.append(record)
            records = new_records
        return insert(table).values(records).returning(*table.columns)

    def get_query(
        self,
        table: FromClause,
        sql_query: QueryType,
        *,
        params: dict | None = None,
    ) -> QueryType:
        """Build an SqlAlchemy query
        :param table: sqlalchemy Table
        :param sql_query: sqlalchemy query type
        :param params: key-value pairs for the query
        :param consumer: optional consumer for manipulating parameters
        """
        filters: list = []
        columns = table.c
        params = params or {}

        for key, value in params.items():
            bits = key.split(":")
            field = bits[0]
            op = bits[1] if len(bits) == 2 else "eq"
            field = getattr(columns, field)
            result = self.default_filter_column(field, op, value)
            if result is not None:
                if not isinstance(result, (list, tuple)):
                    result = (result,)
                filters.extend(result)
        if filters:
            whereclause = and_(*filters) if len(filters) > 1 else filters[0]
            sql_query = cast(Select, sql_query).where(whereclause)
        return sql_query

    def db_count_query(self, sql_query: Select) -> Select:
        return select(func.count()).select_from(sql_query.alias("inner"))

    def order_by_query(
        self,
        table: FromClause,
        sql_query: Select,
        order_by: tuple[str, ...],
    ) -> Select:
        """Apply ordering to a sql_query"""
        return sql_query.order_by(*self.order_by_columns(table, order_by))

    def order_by_columns(
        self,
        table: FromClause,
        order_by: tuple[str, ...],
    ) -> list[Column]:
        """Apply ordering to a sql_query"""
        columns = []
        for name in order_by:
            if name.startswith("-"):
                order_by_column = getattr(table.c, name[1:], None)
                if order_by_column is not None:
                    columns.append(order_by_column.desc())
            else:
                order_by_column = getattr(table.c, name, None)
                if order_by_column is not None:
                    columns.append(order_by_column)
        return columns

    def search_query(
        self,
        table: FromClause,
        sql_query: Select,
        search_fields: tuple[str, ...],
        search: str,
    ) -> Select:
        """Apply search to a sql_query"""
        if search and search_fields:
            columns = [getattr(table.c, col) for col in search_fields]
            return sql_query.where(or_(*(col.ilike(f"%{search}%") for col in columns)))
        return sql_query

    def default_filter_column(self, column: Column, op: str, value: Any) -> Any:
        """
        Applies a filter on a field.
        Notes on 'ne' op:
        Example data: [None, 'john', 'roger']
        ne:john would return only roger (i.e. nulls excluded)
        ne:     would return john and roger
        Notes on  'search' op:
        For some reason, SQLAlchemy uses to_tsquery rather than
        plainto_tsquery for the match operator
        to_tsquery uses operators (&, |, ! etc.) while
        plainto_tsquery tokenises the input string and uses AND between
        tokens, hence plainto_tsquery is what we want here
        For other database back ends, the behaviour of the match
        operator is completely different - see:
        http://docs.sqlalchemy.org/en/rel_1_0/core/sqlelement.html
        :param field: field name
        :param op: 'eq', 'ne', 'gt', 'lt', 'ge', 'le' or 'search'
        :param value: comparison value, string or list/tuple
        :return:
        """
        if multiple := isinstance(value, (list, tuple)):
            value = tuple(column_value_to_python(column, v) for v in value)
        else:
            value = column_value_to_python(column, value)

        if multiple and op in ("eq", "ne"):
            if op == "eq":
                return column.in_(value)
            elif op == "ne":
                return ~column.in_(value)
        else:
            if multiple:
                assert len(value) > 0
                value = value[0]

            if op == "eq":
                return column == value
            elif op == "ne":
                return column != value
            elif op == "gt":
                return column > value
            elif op == "ge":
                return column >= value
            elif op == "lt":
                return column < value
            elif op == "le":
                return column <= value


def column_value_to_python(column: Column, value: Any) -> Any:
    py_type = column.type.python_type
    try:
        if py_type is datetime:
            return parse_date(value) if isinstance(value, str) else value
        elif py_type is date:
            return parse_date(value).date() if isinstance(value, str) else value
        elif py_type is int:
            return int(value)
        else:
            return value
    except Exception as e:
        raise ValidationError("cursor", "invalid cursor") from e



import asyncio
import json
import sys
from importlib import import_module

from fluid.scheduler import TaskManager
from fluid.utils import log

logger = log.get_logger(__name__)


def create_task_manager() -> TaskManager:
    return TaskManager()


async def main(name: str, module_name: str, run_id: str, params: str) -> None:
    log.config()
    module = import_module(module_name)
    task = getattr(module, name)
    kwargs = json.loads(params)
    async with create_task_manager() as task_manager:
        await task_manager.execute(task, run_id=run_id, **kwargs)


if __name__ == "__main__":
    asyncio.get_event_loop().run_until_complete(main(*sys.argv[1:]))



from __future__ import annotations

import asyncio
import enum
import inspect
import json
import logging
import os
import sys
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Callable, Coroutine, NamedTuple, overload

from pydantic import BaseModel, Field, field_serializer
from redis.asyncio.lock import Lock

from fluid import settings
from fluid.utils import kernel, log
from fluid.utils.data import compact_dict
from fluid.utils.dates import as_utc
from fluid.utils.text import create_uid, trim_docstring

from .crontab import Scheduler
from .errors import TaskDecoratorError, TaskRunError

if TYPE_CHECKING:
    from .consumer import TaskManager


TaskExecutor = Callable[["TaskRun"], Coroutine[Any, Any, Any]]
RandomizeType = Callable[[], float | int]


class TaskPriority(enum.StrEnum):
    high = enum.auto()
    medium = enum.auto()
    low = enum.auto()


class TaskState(enum.StrEnum):
    init = enum.auto()
    queued = enum.auto()
    running = enum.auto()
    success = enum.auto()
    failure = enum.auto()
    aborted = enum.auto()
    rate_limited = enum.auto()

    @property
    def is_failure(self) -> bool:
        return self is TaskState.failure

    @property
    def is_done(self) -> bool:
        return self in FINISHED_STATES


FINISHED_STATES = frozenset(
    (TaskState.success, TaskState.failure, TaskState.aborted, TaskState.rate_limited)
)


class TaskManagerConfig(BaseModel):
    schedule_tasks: bool = True
    consume_tasks: bool = True
    max_concurrent_tasks: int = settings.MAX_CONCURRENT_TASKS
    """number of coroutine workers"""
    sleep: float = 0.1
    """amount to sleep after completion of a task"""
    broker_url: str = ""


class TaskInfoBase(BaseModel):
    name: str = Field(description="Task name")
    description: str = Field(description="Task description")
    module: str = Field(description="Task module")
    priority: TaskPriority = Field(description="Task priority")
    schedule: str | None = Field(default=None, description="Task schedule")


class TaskInfoUpdate(BaseModel):
    enabled: bool = Field(default=True, description="Task enabled")
    last_run_end: datetime | None = Field(
        default=None, description="Task last run end as milliseconds since epoch"
    )
    last_run_duration: timedelta | None = Field(
        default=None, description="Task last run duration in milliseconds"
    )
    last_run_state: TaskState | None = Field(
        default=None, description="State of last task run"
    )


class TaskInfo(TaskInfoBase, TaskInfoUpdate):
    pass


class QueuedTask(BaseModel):
    """A task to be queued"""

    run_id: str = Field(description="Task run id")
    task: str = Field(description="Task name")
    params: dict[str, Any] = Field(description="Task parameters")
    priority: TaskPriority | None = Field(default=None, description="Task priority")


class Task(NamedTuple):
    """A Task executes any time it is invoked"""

    name: str
    """Task name - unique identifier"""
    executor: TaskExecutor
    """Task executor function"""
    logger: logging.Logger
    """Task logger"""
    module: str = ""
    """Task python module"""
    short_description: str = ""
    """Short task description - one line"""
    description: str = ""
    """Task description - obtained from the executor docstring if not provided"""
    schedule: Scheduler | None = None
    """Task schedule - None means the task is not scheduled"""
    randomize: RandomizeType | None = None
    params_model: type[BaseModel] | None = None
    max_concurrency: int = 0
    """how many tasks can run in each consumer concurrently - 0 means no limit"""
    timeout_seconds: int = 60
    """Task timeout in seconds - how long the task can run before being aborted"""
    priority: TaskPriority = TaskPriority.medium
    """Task priority - high, medium, low"""

    @property
    def cpu_bound(self) -> bool:
        """True if the task is CPU bound"""
        return self.executor is run_in_subprocess

    def params_dump_json(self, params: dict[str, Any]) -> str:
        """Dumps task parameters as JSON string"""
        if params_model := self.params_model:
            return params_model(**params).model_dump_json()
        return json.dumps(params)

    def info(self, **params: Any) -> TaskInfo:
        params.update(
            name=self.name,
            description=self.description,
            module=self.module,
            priority=self.priority,
            schedule=str(self.schedule) if self.schedule else None,
        )
        return TaskInfo(**compact_dict(params))


class TaskRun(BaseModel, arbitrary_types_allowed=True):
    """A TaskRun contains all the data generated by a Task run

    This model is never initialized directly, it is created by the TaskManager
    """

    id: str
    task: Task
    priority: TaskPriority
    params: dict[str, Any]
    state: TaskState = TaskState.init
    task_manager: TaskManager = Field(exclude=True, repr=False)
    queued: datetime | None = None
    start: datetime | None = None
    end: datetime | None = None

    async def execute(self) -> None:
        try:
            self.set_state(TaskState.running)
            await self.task.executor(self)
        except Exception:
            self.set_state(TaskState.failure)
            raise
        else:
            self.set_state(TaskState.success)

    @field_serializer("task")
    def serialize_task(self, task: Task, _info: Any) -> str:
        return task.name

    @property
    def logger(self) -> logging.Logger:
        return self.task.logger

    @property
    def in_queue(self) -> timedelta | None:
        if self.queued and self.start:
            return self.start - self.queued
        return None

    @property
    def duration(self) -> timedelta | None:
        if self.start and self.end:
            return self.end - self.start
        return None

    @property
    def duration_ms(self) -> float | None:
        duration = self.duration
        if duration is not None:
            return round(1000 * duration.total_seconds(), 2)
        return None

    @property
    def total(self) -> timedelta | None:
        if self.queued and self.end:
            return self.end - self.queued
        return None

    @property
    def name(self) -> str:
        return self.task.name

    @property
    def name_id(self) -> str:
        return f"{self.task.name}.{self.id}"

    @property
    def is_done(self) -> bool:
        return self.state.is_done

    @property
    def is_failure(self) -> bool:
        return self.state.is_failure

    def params_dump_json(self) -> str:
        return self.task.params_dump_json(self.params)

    def set_state(
        self,
        state: TaskState,
        state_time: datetime | None = None,
    ) -> None:
        if self.state == state:
            return
        state_time = as_utc(state_time)
        match (self.state, state):
            case (TaskState.init, TaskState.queued):
                self.queued = state_time
                self.state = state
                self._dispatch()
            case (TaskState.init, _):
                self.set_state(TaskState.queued, state_time)
                self.set_state(state, state_time)
            case (TaskState.queued, TaskState.running):
                self.start = state_time
                self.state = state
                self._dispatch()
            case (
                TaskState.queued,
                TaskState.success
                | TaskState.aborted
                | TaskState.rate_limited
                | TaskState.failure,
            ):
                self.set_state(TaskState.running, state_time)
                self.set_state(state, state_time)
            case (
                TaskState.running,
                TaskState.success
                | TaskState.aborted
                | TaskState.rate_limited
                | TaskState.failure,
            ):
                self.end = state_time
                self.state = state
                self._dispatch()
            case _:
                raise TaskRunError(f"invalid state transition {self.state} -> {state}")

    def lock(self, timeout: float | None) -> Lock:
        return self.task_manager.broker.lock(self.name, timeout=timeout)

    def _dispatch(self) -> None:
        self.task_manager.dispatcher.dispatch(self.model_copy())


@dataclass
class TaskRunWaiter:
    task_manager: TaskManager
    uid: str = field(default_factory=create_uid)
    _runs: dict[str, TaskRun] = field(default_factory=dict)

    def __enter__(self) -> TaskRunWaiter:
        for state in FINISHED_STATES:
            self.task_manager.dispatcher.register_handler(f"{state}.{self.uid}", self)
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        for state in FINISHED_STATES:
            self.task_manager.dispatcher.unregister_handler(f"{state}.{self.uid}")

    def __call__(self, task_run: TaskRun) -> None:
        self._runs[task_run.id] = task_run

    async def wait(self, task_run: TaskRun, *, timeout: int = 2) -> TaskRun:
        async with asyncio.timeout(timeout):
            while task_run.id not in self._runs:
                await asyncio.sleep(0.01)
        return self._runs[task_run.id]


@overload
def task(executor: TaskExecutor) -> Task: ...


@overload
def task(
    *,
    name: str | None = None,
    schedule: Scheduler | None = None,
    short_description: str | None = None,
    description: str | None = None,
    randomize: RandomizeType | None = None,
    max_concurrency: int = 1,
    priority: TaskPriority = TaskPriority.medium,
    params_model: BaseModel | None = None,
    cpu_bound: bool = False,
    timeout_seconds: int = 60,
) -> TaskConstructor: ...


# implementation of the task decorator
def task(executor: TaskExecutor | None = None, **kwargs: Any) -> Task | TaskConstructor:
    """Decorator to create a Task

    This decorator can be used in two ways:

    - As a simple decorator of the executor function
    - As a function with keyword arguments
    """
    if kwargs and executor:
        raise TaskDecoratorError("cannot use positional parameters")
    elif kwargs:
        return TaskConstructor(**kwargs)
    elif not executor:
        raise TaskDecoratorError("this is a decorator cannot be invoked in this way")
    else:
        return TaskConstructor()(executor)


class TaskConstructor:
    def __init__(self, *, cpu_bound: bool = False, **kwargs: Any) -> None:
        self.cpu_bound = cpu_bound
        self.kwargs = kwargs

    def __call__(self, executor: TaskExecutor) -> Task:
        if self.cpu_bound:
            return self.cpu_bound_task(executor)
        else:
            return self.create_task(executor)

    def create_task(
        self,
        executor: TaskExecutor,
        defaults: dict[str, Any] | None = None,
    ) -> Task:
        kwargs: dict[str, Any] = self.kwargs_defaults(executor)
        if defaults:
            kwargs.update(defaults)
        kwargs.update(self.kwargs)
        name = kwargs["name"]
        kwargs.update(
            executor=executor,
            logger=log.get_logger(f"task.{name}", prefix=True),
        )
        return Task(**kwargs)

    def cpu_bound_task(self, executor: TaskExecutor) -> Task:
        if is_in_subprocess():
            return self.create_task(executor)
        else:
            return self.create_task(run_in_subprocess, self.kwargs_defaults(executor))

    def kwargs_defaults(self, executor: TaskExecutor) -> dict[str, Any]:
        module = inspect.getmodule(executor)
        description = trim_docstring(inspect.getdoc(executor) or "")
        short_description = description.split("\n")[0].strip()
        return dict(
            name=get_name(executor),
            module=module.__name__ if module else "",
            description=description,
            short_description=short_description,
            executor=executor,
        )


def get_name(o: Any) -> str:
    if hasattr(o, "__name__"):
        return str(o.__name__)
    elif hasattr(o, "__class__"):
        return str(o.__class__.__name__)
    else:
        return str(o)


def is_in_subprocess() -> bool:
    return os.getenv("TASK_MANAGER_SPAWN") == "true"


class RemoteLog:
    def __init__(self, out: Any) -> None:
        self.out = out

    def __call__(self, data: bytes) -> None:
        self.out.write(data.decode("utf-8"))


async def run_in_subprocess(ctx: TaskRun) -> None:
    env = dict(os.environ)
    env["TASK_MANAGER_SPAWN"] = "true"
    result = await kernel.run_python(
        "-W",
        "ignore",
        "-m",
        "fluid.scheduler.cpubound",
        ctx.name,
        ctx.task.module,
        ctx.id,
        ctx.task.params_dump_json(ctx.params),
        result_callback=RemoteLog(sys.stdout),
        error_callback=RemoteLog(sys.stderr),
        env=env,
        stream_output=True,
        stream_error=True,
    )
    if result:
        raise TaskRunError(result)



from __future__ import annotations

from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, cast

import sqlalchemy as sa
from alembic import command as alembic_cmd
from alembic.config import Config
from sqlalchemy.engine import Engine
from sqlalchemy_utils import create_database, database_exists

if TYPE_CHECKING:
    from .container import Database


class Migration:
    def __init__(self, db: Database) -> None:
        self.db = db
        self.cfg = create_config(db)

    @property
    def metadata(self) -> sa.MetaData:
        return self.db.metadata

    @property
    def sync_engine(self) -> Engine:
        return self.db.sync_engine

    def init(self) -> str:
        dirname = self.cfg.get_main_option("script_location") or ""
        alembic_cmd.init(self.cfg, dirname)
        return self.message()

    def show(self, revision: str) -> str:
        alembic_cmd.show(self.cfg, revision)
        return self.message()

    def history(self) -> str:
        alembic_cmd.history(self.cfg)
        return self.message()

    def revision(
        self,
        message: str,
        autogenerate: bool = False,
        branch_label: str | None = None,
    ) -> str:
        alembic_cmd.revision(
            self.cfg,
            autogenerate=autogenerate,
            message=message,
            branch_label=branch_label,
        )
        return self.message()

    def upgrade(self, revision: str) -> str:
        alembic_cmd.upgrade(self.cfg, revision)
        return self.message()

    def downgrade(self, revision: str) -> str:
        alembic_cmd.downgrade(self.cfg, revision)
        return self.message()

    def current(self, verbose: bool = False) -> str:
        alembic_cmd.current(self.cfg, verbose=verbose)
        return self.message()

    def message(self) -> str:
        msg = cast(StringIO, self.cfg.stdout).getvalue()
        self.cfg.stdout.seek(0)
        self.cfg.stdout.truncate()
        return msg

    def db_exists(self, dbname: str = "") -> bool:
        url = self.sync_engine.url
        if dbname:
            url = url.set(database=dbname)
        return database_exists(url)

    def db_create(self, dbname: str = "") -> bool:
        """Creates a new database if it does not exist"""
        url = self.sync_engine.url
        if dbname:
            url = url.set(database=dbname)
        if database_exists(url):
            return False
        create_database(url)
        return True

    def truncate_all(self) -> None:
        """Drop all tables from :attr:`metadata` in database"""
        with self.sync_engine.begin() as conn:
            conn.execute(sa.text(f'truncate {", ".join(self.metadata.tables)}'))

    def drop_all_schemas(self) -> None:
        """Drop all schema in database"""
        with self.sync_engine.begin() as conn:
            conn.execute(sa.text("DROP SCHEMA IF EXISTS public CASCADE"))
            conn.execute(sa.text("CREATE SCHEMA IF NOT EXISTS public"))

    def create_ro_user(
        self,
        username: str,
        password: str,
        role: str = "",
        schema: str = "public",
    ) -> bool:
        """Creates a read-only user"""
        engine = self.sync_engine
        role = role or f"{engine.url.username}_ro"
        database = engine.url.database
        created = True
        with engine.begin() as conn:
            try:
                conn.execute(sa.text(f"CREATE ROLE {role};"))
            except sa.exc.ProgrammingError:
                created = False
        with engine.begin() as conn:
            conn.execute(
                sa.text(
                    f"GRANT CONNECT ON DATABASE {database} TO {role};"
                    f"GRANT USAGE ON SCHEMA {schema} TO {role};"
                    f"GRANT SELECT ON ALL TABLES IN SCHEMA {schema} TO {role};"
                    f"GRANT SELECT ON ALL SEQUENCES IN SCHEMA {schema} TO {role};",
                ),
            )
            conn.execute(
                sa.text(
                    f"ALTER DEFAULT PRIVILEGES IN SCHEMA {schema} "
                    f"GRANT SELECT ON TABLES TO {role};",
                ),
            )
        with engine.begin() as conn:
            try:
                conn.execute(
                    sa.text(
                        f"CREATE USER {username} WITH PASSWORD '{password}';"
                        f"GRANT {role} TO {username};",
                    ),
                )
            except sa.exc.ProgrammingError:
                created = False
        return created

    def drop_role(
        self,
        role: str,
    ) -> bool:
        """Drop a role"""
        try:
            with self.sync_engine.begin() as conn:
                conn.execute(sa.text(f"DROP OWNED BY {role};"))
            with self.sync_engine.begin() as conn:
                conn.execute(sa.text(f"DROP ROLE IF EXISTS {role};"))
        except sa.exc.ProgrammingError as exc:
            if f'role "{role}" does not exist' not in str(exc):
                raise
            return False
        return True


def create_config(db: "Database") -> Config:
    """Programmatically create Alembic config"""
    cfg = Config(stdout=StringIO())
    cfg.set_main_option("script_location", str(db.migration_path))
    cfg.config_file_name = str(Path(db.migration_path) / "alembic.ini")
    cfg.set_section_option(
        "alembic",
        "sqlalchemy.url",
        db.sync_engine.url.render_as_string(hide_password=False),
    )
    # create empty logging section to avoid raising errors in env.py
    cfg.set_section_option("logging", "path", "")
    cfg.metadata = db.metadata  # type: ignore
    return cfg



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

import click
import uvicorn
from fastapi import FastAPI
from rich.console import Console
from rich.table import Table
from uvicorn.importer import import_from_string

from fluid.utils import log
from fluid.utils.lazy import LazyGroup

if TYPE_CHECKING:
    from .consumer import TaskManager
    from .models import TaskRun


TaskManagerApp = FastAPI | Callable[..., Any] | str


class TaskManagerCLI(LazyGroup):
    def __init__(
        self,
        task_manager_app: TaskManagerApp,
        **kwargs: Any,
    ):
        kwargs.setdefault("commands", DEFAULT_COMMANDS)
        super().__init__(**kwargs)
        self.task_manager_app = task_manager_app


def ctx_task_manager_app(ctx: click.Context) -> TaskManagerApp:
    return ctx.parent.command.task_manager_app  # type: ignore


def ctx_app(ctx: click.Context) -> FastAPI:
    app = ctx_task_manager_app(ctx)  # type: ignore
    if isinstance(app, str):
        return import_from_string(app)()
    elif isinstance(app, FastAPI):
        return app
    else:
        return app()


def ctx_task_manager(ctx: click.Context) -> TaskManager:
    return ctx_app(ctx).state.task_manager


@click.command()
@click.pass_context
def ls(ctx: click.Context) -> None:
    """list all tasks"""
    task_manager = ctx_task_manager(ctx)
    table = Table(title="Tasks")
    table.add_column("Name", style="cyan", no_wrap=True)
    table.add_column("Schedule", style="magenta")
    table.add_column("CPU bound", style="magenta")
    table.add_column("Timeout secs", style="green")
    table.add_column("Priority", style="magenta")
    table.add_column("Description", style="green")
    for name in sorted(task_manager.registry):
        task = task_manager.registry[name]
        table.add_row(
            name,
            str(task.schedule),
            "yes" if task.cpu_bound else "no",
            str(task.timeout_seconds),
            str(task.priority),
            task.short_description,
        )
    console = Console()
    console.print(table)


@click.command()
@click.pass_context
@click.argument("task")
@click.option(
    "--dry-run",
    is_flag=True,
    help="dry run (if the tasks supports it)",
    default=False,
)
def execute(ctx: click.Context, task: str, dry_run: bool) -> None:
    """execute a task"""
    task_manager = ctx_task_manager(ctx)
    run = task_manager.execute_sync(task, dry_run=dry_run)
    console = Console()
    console.print(task_run_table(run))


@click.command("serve", short_help="Start app server.")
@click.option(
    "--host",
    "-h",
    default="0.0.0.0",
    help="The interface to bind to",
    show_default=True,
)
@click.option(
    "--port",
    "-p",
    default=8080,
    help="The port to bind to",
    show_default=True,
)
@click.option(
    "--reload",
    is_flag=True,
    default=False,
    help="Enable auto-reload",
    show_default=True,
)
@click.pass_context
def serve(ctx: click.Context, host: str, port: int, reload: bool) -> None:
    """Run the service."""
    task_manager_app = ctx_task_manager_app(ctx)
    uvicorn.run(
        task_manager_app,
        port=port,
        host=host,
        log_level="info",
        reload=reload,
        log_config=log.config(),
    )


DEFAULT_COMMANDS = (ls, execute, serve)


def task_run_table(task_run: TaskRun) -> Table:
    table = Table(title="Task Run", show_header=False)
    color = "red" if task_run.state.is_failure else "green"
    table.add_column("Name", style="cyan")
    table.add_column("Description", style=color)
    table.add_row("name", task_run.task.name)
    table.add_row("description", task_run.task.description)
    table.add_row("run_id", task_run.id)
    table.add_row("state", task_run.state)
    if task_run.start:
        table.add_row("started", task_run.start.isoformat())
    if task_run.end:
        table.add_row("completed", task_run.end.isoformat())
    table.add_row("duration ms", str(task_run.duration_ms))
    return table



from typing import Any, Hashable, Iterable

DEFAULT_SKIP_VALUES = frozenset((None,))


def reversed_dict(d: dict) -> dict:  # type: ignore [type-arg]
    return {v: k for k, v in d.items()}


def compact_dict(
    *args: Iterable[Any],
    skip_values: set[Any] | frozenset[Any] = DEFAULT_SKIP_VALUES,
    **kwargs: Any,
) -> dict[str, Any]:
    return {
        k: v
        for k, v in dict(*args, **kwargs).items()
        if isinstance(v, bool) or not isinstance(v, Hashable) or v not in skip_values
    }



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

import click
from sqlalchemy import func, inspect, select, text

if TYPE_CHECKING:
    from .container import Database


class DbGroup(click.Group):
    def __init__(
        self,
        db: Database,
        name: str = "db",
        help: str = "Manage database and migrations",  # noqa: A002
        **kwargs: Any,
    ) -> None:
        super().__init__(name=name, help=help, **kwargs)
        self.db = db
        for command in _db.commands.values():
            self.add_command(command)

    def get_command(self, ctx: click.Context, name: str) -> Optional[click.Command]:
        ctx.obj = {"db": self.db}
        return super().get_command(ctx, name)

    def list_commands(self, ctx: click.Context) -> list[str]:
        ctx.obj = {"db": self.db}
        return super().list_commands(ctx)


def get_db(ctx: click.Context) -> Database:
    return ctx.obj["db"]


@click.group()
def _db() -> None:
    """Perform database migrations and utilities"""
    pass


@_db.command()
@click.pass_context
def dsn(ctx: click.Context) -> None:
    """Display data source name"""
    click.echo(get_db(ctx).engine.url.render_as_string(hide_password=False))


@_db.command()
@click.pass_context
def init(ctx: click.Context) -> None:
    """Creates a new migration repository."""
    get_db(ctx).migration().init()


@_db.command()
@click.option("-m", "--message", help="Revision message", required=True)
@click.option(
    "--branch-label",
    help="Specify a branch label to apply to the new revision",
)
@click.pass_context
def migrate(ctx: click.Context, message: str, branch_label: str) -> None:
    """Autogenerate a new revision file
    alias for 'revision --autogenerate'
    """
    get_db(ctx).migration().revision(
        message, autogenerate=True, branch_label=branch_label
    )


@_db.command()
@click.option("-m", "--message", help="Revision message", required=True)
@click.option(
    "--branch-label",
    help="Specify a branch label to apply to the new revision",
)
@click.option(
    "--autogenerate",
    default=False,
    is_flag=True,
    help=(
        "Populate revision script with candidate migration "
        "operations, based on comparison of database to model"
    ),
)
@click.pass_context
def revision(
    ctx: click.Context,
    message: str,
    branch_label: str,
    autogenerate: bool,
) -> None:
    """Autogenerate a new revision file"""
    get_db(ctx).migration().revision(
        message,
        autogenerate=autogenerate,
        branch_label=branch_label,
    )


@_db.command()
@click.argument("revision", default="heads")
@click.pass_context
def upgrade(ctx: click.Context, revision: Union[str, int]) -> None:
    """Upgrade to a later version"""
    get_db(ctx).migration().upgrade(str(revision))
    click.echo(f"upgraded successfully to {revision}")


@_db.command()
@click.argument("revision")
@click.pass_context
def downgrade(ctx: click.Context, revision: Union[str, int]) -> None:
    """Downgrade to a previous REVISION identifier"""
    get_db(ctx).migration().downgrade(str(revision))
    click.echo(f"downgraded successfully to {revision}")


@_db.command()
@click.option("--revision", default="heads")
@click.pass_context
def show(ctx: click.Context, revision: str) -> None:
    """Show revision ID and creation date"""
    click.echo(get_db(ctx).migration().show(revision))


@_db.command()
@click.pass_context
def history(ctx: click.Context) -> None:
    """List changeset scripts in chronological order"""
    click.echo(get_db(ctx).migration().history())


@_db.command()
@click.option("--verbose/--quiet", default=False)
@click.pass_context
def current(ctx: click.Context, verbose: bool) -> None:
    """Show revision ID and creation date"""
    click.echo(get_db(ctx).migration().current(verbose))


@_db.command()
@click.argument("dbname", nargs=1)
@click.pass_context
def create(ctx: click.Context, dbname: str) -> None:
    """Creates a new database"""
    if get_db(ctx).migration().db_create(dbname):
        click.echo(f"database {dbname} created")
    else:
        click.echo(f"database {dbname} already available")


@_db.command()
@click.argument("name", nargs=1)
@click.pass_context
def create_schema(ctx: click.Context, name: str) -> None:
    """Creates a new database schema"""
    with get_db(ctx).sync_engine.connect() as conn:
        conn.execute(text(f"create schema {name}"))
    click.echo(f"schema {name} created")


@_db.command()
@click.argument("username", nargs=1)
@click.argument("password", nargs=1)
@click.option("--dbname", help="database name")
@click.option("--role", help="role name")
@click.option("--schema", default="public", show_default=True)
@click.pass_context
def create_ro_user(
    ctx: click.Context,
    username: str,
    password: str,
    dbname: str,
    role: str,
    schema: str,
) -> None:
    """Creates or updates a database user"""
    d = get_db(ctx)
    if dbname:
        url = d.engine.url.set(database=dbname)
        d = type(d).from_env(dsn=str(url))
    mig = d.migration()
    mig.create_ro_user(username, password, role=role, schema=schema)
    click.echo(f"schema {username} created")


@_db.command()
@click.argument("role", nargs=1)
@click.pass_context
def drop_role(ctx: click.Context, role: str) -> None:
    """Drop a role or a user"""
    mig = get_db(ctx).migration()
    if mig.drop_role(role):
        click.echo(f"role {role} dropped")
    else:
        click.echo(f"role {role} not found")


@_db.command()
@click.option(
    "--db",
    default=False,
    is_flag=True,
    help="List tables in database rather than in sqlalchemy metadata",
)
@click.pass_context
def tables(ctx: click.Context, db: bool) -> None:
    """List all tables managed by the app"""
    d = get_db(ctx)
    if db:
        tables = inspect(d.sync_engine).get_table_names()
    else:
        tables = list(d.tables)
    for name in sorted(tables):
        click.echo(name)


@_db.command()
@click.pass_context
def databases(ctx: click.Context) -> None:
    """List databases"""
    sql = text("SELECT datname FROM pg_database WHERE datistemplate = false;")
    engine = get_db(ctx).sync_engine
    with engine.begin() as conn:
        for row in conn.execute(sql):
            click.echo(f"{row[0]}")


@_db.command()
@click.argument("table", nargs=1)
@click.option("--dry", help="dry run", is_flag=True, default=False)
@click.pass_context
def delete_rows(ctx: click.Context, table: str, dry: bool) -> None:
    """Delete all rows in a table"""
    db = get_db(ctx)
    engine = db.sync_engine
    try:
        db_table = db.metadata.tables[table]
    except KeyError:
        click.echo(f"table {table} not found in {engine}", err=True)
        raise click.Abort() from None
    with engine.begin() as conn:
        count_query = select(func.count()).select_from(db_table)
        rows = conn.execute(count_query).scalar()
        click.echo(f"removing {rows} rows")
        if dry:
            click.echo("nothing done, dry mode")
        else:
            conn.execute(db_table.delete())



from functools import partial
from typing import Any, Sequence, cast

import sqlalchemy as sa
from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo


def model_from_table(
    model_name: str,
    table: sa.Table,
    *,
    exclude: Sequence[str] | None = None,
    include: Sequence[str] | None = None,
    default: bool | Sequence[str] = False,
    required: bool | Sequence[str] = False,
    ops: dict[str, Sequence[str]] | None = None,
) -> type[BaseModel]:
    """Create a dataclass from an :class:`sqlalchemy.schema.Table`

    :param name: dataclass name
    :param table: sqlalchemy table
    :param exclude: fields to exclude from the dataclass
    :param include: fields to include in the dataclass
    :param default: use columns defaults in the dataclass
    :param required: set non nullable columns without a default as
        required fields in the dataclass
    :param ops: additional operation for fields
    """
    includes = set(include or table.columns.keys()) - set(exclude or ())
    default_set = column_info(includes, default)
    required_set = column_info(includes, required)
    column_ops = cast(dict[str, Sequence[str]], ops or {})
    fields = {}
    for column in table.columns:
        name = str(column.name)
        if name not in includes:
            continue
        python_type: type | None = None
        if hasattr(column.type, "impl"):
            if hasattr(column.type.impl, "python_type"):
                python_type = column.type.impl.python_type
        elif hasattr(column.type, "python_type"):
            python_type = column.type.python_type
        if not python_type:  # pragma:   no cover
            raise NotImplementedError(f"Cannot convert column {name}")
        required = name in required_set
        use_default = name in default_set
        fields[name] = (
            python_type,
            field_info(column, required, use_default, column_ops.get(name, ())),
        )
    return create_model(model_name, **fields)  # type: ignore


def column_info(columns: set[str], value: bool | Sequence[str]) -> set[str]:
    if value is False:
        return set()
    elif value is True:
        return columns.copy()
    else:
        return set(value if value is not None else columns)


def field_info(
    col: sa.Column, required: bool, use_default: bool, ops: Sequence[str]
) -> FieldInfo:
    data: dict[str, Any] = {"ops": ops}
    data.clear()
    if use_default and col.default is not None:
        default = col.default.arg  # type: ignore
        if col.default.is_callable:
            data.update(default_factory=partial(default, None))
            required = False
        elif col.default.is_sequence:
            data.update(default_factory=lambda: default.copy())
            required = False
        else:
            data.update(default=default)
            if required and (col.nullable or default is not None):
                required = False
    elif required and col.nullable:
        required = False
    # data.update(required=required)
    if col.doc:
        data.update(description=col.doc)
    data.update(col.info or ())
    return FieldInfo(**data)



from __future__ import annotations

import asyncio
import logging
import random
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass, field
from typing import (
    AsyncGenerator,
    Awaitable,
    Callable,
    Generator,
    Iterator,
    Protocol,
    Sequence,
    TypeVar,
)

from inflection import underscore

from fluid import settings

from .dispatcher import AsyncDispatcher, MessageType

logger = logging.getLogger(__name__)


class HasWorkers(ABC):
    @abstractmethod
    def get_workers(self) -> Sequence[Worker]:
        """Get the workers that can be added to a the Workers Manager"""


class Worker(ABC):
    """The base class of a worker that can be run"""

    def __init__(self, name: str = "") -> None:
        self._name: str = name or underscore(type(self).__name__)

    @property
    def worker_name(self) -> str:
        """The name of the worker"""
        return self._name

    @property
    def num_workers(self) -> int:
        """The number of workers in this worker"""
        return 1

    @abstractmethod
    async def status(self) -> dict:
        """
        Get the status of the worker.
        """

    @abstractmethod
    def gracefully_stop(self) -> None:
        "gracefully stop the worker"

    @abstractmethod
    def is_running(self) -> bool:
        """Is the worker running?"""

    @abstractmethod
    def is_stopping(self) -> bool:
        """Is the worker stopping?"""

    @abstractmethod
    async def run(self) -> None:
        """run the worker"""


class RunningWorker(Worker):
    """A Worker that can be started"""

    def __init__(self, name: str = "") -> None:
        super().__init__(name)
        self._running: bool = False

    def is_running(self) -> bool:
        return self._running

    @contextmanager
    def start_running(self) -> Generator:
        if self._running:
            raise RuntimeError("Worker is already running")
        self._running = True
        try:
            logger.info("%s started running", self.worker_name)
            yield
        finally:
            self._running = False
            logger.warning("%s stopped running", self.worker_name)


class StoppingWorker(RunningWorker):
    """A Worker that can be stopped"""

    def __init__(self, name: str = "") -> None:
        super().__init__(name)
        self._stopping: bool = False

    def is_stopping(self) -> bool:
        return self._stopping

    def gracefully_stop(self) -> None:
        self._stopping = True

    async def status(self) -> dict:
        return {"stopping": self.is_stopping(), "running": self.is_running()}


class WorkerFunction(StoppingWorker):
    """A Worker that runs a coroutine function"""

    def __init__(
        self,
        run_function: Callable[[], Awaitable[None]],
        heartbeat: float | int = 0,
        name: str = "",
    ) -> None:
        super().__init__(name=name)
        self._run_function = run_function
        self._heartbeat = heartbeat

    async def run(self) -> None:
        with self.start_running():
            while not self.is_stopping():
                await self._run_function()
                await asyncio.sleep(self._heartbeat)


T = TypeVar("T", contravariant=True)


class MessageProducer(Protocol[T]):
    def send(self, message: T | None) -> None: ...


class QueueConsumer(StoppingWorker, MessageProducer[MessageType]):
    """A Worker that can receive messages

    This worker can receive messages but not consume them.
    """

    def __init__(self, name: str = "") -> None:
        super().__init__(name=name)
        self._queue: asyncio.Queue[MessageType | None] = asyncio.Queue()

    async def get_message(self, timeout: float = 0.5) -> MessageType | None:
        try:
            async with asyncio.timeout(timeout):
                return await self._queue.get()
        except asyncio.TimeoutError:
            return None
        except (asyncio.CancelledError, RuntimeError):
            if not self.is_stopping():
                raise
        return None

    def queue_size(self) -> int:
        return self._queue.qsize()

    async def status(self) -> dict:
        status = await super().status()
        status.update(queue_size=self.queue_size())
        return status

    def send(self, message: MessageType | None) -> None:
        """Send a message into the worker"""
        self._queue.put_nowait(message)


class QueueConsumerWorker(QueueConsumer[MessageType]):
    """A Worker that can receive and consume messages"""

    def __init__(
        self,
        on_message: Callable[[MessageType], Awaitable[None]],
        name: str = "",
    ) -> None:
        super().__init__(name=name)
        self.on_message = on_message

    async def run(self) -> None:
        with self.start_running():
            while not self.is_stopping():
                message = await self.get_message()
                if message is not None:
                    await self.on_message(message)


class AsyncConsumer(QueueConsumer[MessageType]):
    """A Worker that can dispatch async callbacks"""

    AsyncCallback: AsyncDispatcher[MessageType]

    def __init__(
        self, dispatcher: AsyncDispatcher[MessageType], name: str = ""
    ) -> None:
        super().__init__(name)
        self.dispatcher: AsyncDispatcher[MessageType] = dispatcher

    async def run(self) -> None:
        with self.start_running():
            while not self.is_stopping():
                message = await self.get_message()
                if message is not None:
                    await self.dispatcher.dispatch(message)


@dataclass
class WorkerTasks:
    workers: Sequence[Worker] = field(default_factory=list)
    tasks: Sequence[asyncio.Task] = field(default_factory=list)

    def is_stopping(self) -> bool:
        return any(worker.is_stopping() for worker in self.workers)

    @property
    def num_workers(self) -> int:
        return sum(worker.num_workers for worker in self.workers)

    @property
    def frozen(self) -> bool:
        if isinstance(self.workers, list) and isinstance(self.tasks, list):
            return False
        return True

    def get_worker_by_name(self, worker_name: str) -> Worker | None:
        for worker in self.workers:
            if worker.worker_name == worker_name:
                return worker
        return None

    def workers_tasks(self) -> tuple[list[Worker], list[asyncio.Task]]:
        if self.frozen:
            raise RuntimeError("Cannot add workers")
        return self.workers, self.tasks  # type: ignore

    def gracefully_stop(self) -> None:
        for worker in self.workers:
            worker.gracefully_stop()

    async def status(self) -> dict:
        status_workers = await asyncio.gather(
            *[worker.status() for worker in self.workers],
        )
        return {
            worker.worker_name: status
            for worker, status in zip(self.workers, status_workers)
        }

    def cancel(self) -> None:
        for task in self.tasks:
            task.cancel()


class MultipleWorkers(RunningWorker):
    """A worker managing several workers"""

    def __init__(
        self,
        *workers: Worker,
        name: str = "",
        heartbeat: float | int = 0.1,
        stopping_grace_period: int = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(name)
        self._heartbeat = heartbeat
        self._workers = WorkerTasks()
        self._has_shutdown = False
        self._force_shutdown = False
        self._stopping_grace_period = stopping_grace_period
        self.add_workers(*workers)

    @abstractmethod
    def add_workers(self, *workers: Worker) -> None: ...

    @abstractmethod
    async def wait_for_exit(self) -> None: ...

    def is_stopping(self) -> bool:
        return self._workers.is_stopping()

    @property
    def num_workers(self) -> int:
        return self._workers.num_workers

    def __iter__(self) -> Iterator[Worker]:
        return iter(self._workers.workers)

    def gracefully_stop(self) -> None:
        self._workers.gracefully_stop()

    async def status(self) -> dict:
        return await self._workers.status()

    def create_task(self, worker: Worker) -> asyncio.Task:
        return asyncio.create_task(
            self._run_worker(worker), name=f"{self.worker_name}-{worker.worker_name}"
        )

    async def on_shutdown(self) -> None:
        """called after the workers are stopped"""

    async def shutdown(self) -> None:
        """shutdown the workers"""
        if self._has_shutdown:
            return
        self._has_shutdown = True
        logger.warning(
            "gracefully stopping %d workers: %s",
            self.num_workers,
            ", ".join(w.worker_name for w in self._workers.workers),
        )
        self.gracefully_stop()
        try:
            async with asyncio.timeout(self._stopping_grace_period):
                await self.wait_for_exit()
            await self.on_shutdown()
            return
        except asyncio.TimeoutError:
            logger.warning(
                "could not stop workers %s gracefully after %s"
                " seconds - force shutdown",
                ", ".join(
                    task.get_name() for task in self._workers.tasks if not task.done()
                ),
                self._stopping_grace_period,
            )
        except asyncio.CancelledError:
            pass
        self._force_shutdown = True
        self._workers.cancel()
        try:
            await self.wait_for_exit()
        except asyncio.CancelledError:
            pass
        await self.on_shutdown()

    def bail_out(self, reason: str, code: int = 1) -> None:
        self.gracefully_stop()

    @asynccontextmanager
    async def safe_run(self) -> AsyncGenerator:
        """Context manager to run a worker safely"""
        try:
            yield
        except asyncio.CancelledError:
            if self._force_shutdown:
                # we are shutting down, this is expected
                pass
            raise
        except Exception as e:
            reason = f"unhandled exception while running workers: {e}"
            logger.exception(reason)
            asyncio.get_event_loop().call_soon(self.bail_out, reason, 2)
        else:
            # worker finished without error
            # make sure we are shutting down
            asyncio.get_event_loop().call_soon(self.bail_out, "worker exit", 1)

    async def _run_worker(self, worker: Worker) -> None:
        async with self.safe_run():
            await worker.run()


class DynamicWorkers(MultipleWorkers):
    def add_workers(self, *workers: Worker) -> None:
        """add workers to the workers

        They can be added while the workers are running.
        """
        workers_, tasks_ = self._workers.workers_tasks()
        for worker in workers:
            workers_.append(worker)
            tasks_.append(self.create_task(worker))

    async def run(self) -> None:
        with self.start_running():
            while not self.is_stopping():
                for worker, task in zip(self._workers.workers, self._workers.tasks):
                    if worker.is_stopping() or task.done():
                        break
                await asyncio.sleep(self._heartbeat)
            await self.shutdown()

    async def wait_for_exit(self) -> None:
        await asyncio.gather(*self._workers.tasks)


class Workers(MultipleWorkers):
    """A worker managing several workers"""

    def __init__(
        self,
        *workers: Worker,
        name: str = "",
        stopping_grace_period: int = settings.STOPPING_GRACE_PERIOD,
    ) -> None:
        super().__init__(
            *workers, name=name, stopping_grace_period=stopping_grace_period
        )
        self._workers_task: asyncio.Task | None = None
        self._delayed_callbacks: list[
            tuple[Callable[[], None], float, float, float]
        ] = []

    @property
    def running(self) -> bool:
        return self._workers.frozen

    # Multiple Worker implementation
    def add_workers(self, *workers: Worker) -> None:
        """add workers to the workers"""
        workers_, _ = self._workers.workers_tasks()
        for worker in workers:
            if worker not in workers_:
                workers_.append(worker)

    async def run(self) -> None:
        """run the workers"""
        with self.start_running():
            async with self.safe_run():
                workers, _ = self._workers.workers_tasks()
                self._workers.workers = tuple(workers)
                self._workers.tasks = tuple(
                    self.create_task(worker) for worker in workers
                )
                await asyncio.gather(*self._workers.tasks)
            await self.shutdown()

    async def wait_for_exit(self) -> None:
        if self._workers_task is not None:
            await self._workers_task

    def remove_workers(self, *workers: Worker) -> None:
        "remove workers from the workers"
        workers_, _ = self._workers.workers_tasks()
        for worker in workers:
            try:
                workers_.remove(worker)
            except ValueError:
                pass

    async def startup(self) -> None:
        """start the workers"""
        if self._workers_task is None:
            self._workers_task = asyncio.create_task(self.run(), name=self.worker_name)
            for args in self._delayed_callbacks:
                self._delayed_callback(*args)
            self._delayed_callbacks = []

    def register_callback(
        self,
        callback: Callable[[], None],
        seconds: float,
        jitter: float = 0.0,
        periodic: bool | float = False,
    ) -> None:
        """Register a callback

        The callback can be periodic or not.
        """
        if periodic is True:
            periodic_float = seconds
        elif periodic is False:
            periodic_float = 0.0
        else:
            periodic_float = periodic
        if not self.running:
            self._delayed_callbacks.append((callback, seconds, jitter, periodic_float))
        else:
            self._delayed_callback(callback, seconds, jitter, periodic_float)

    def _delayed_callback(
        self,
        callback: Callable[[], None],
        delay: float,
        jitter: float,
        periodic: float,
    ) -> None:
        asyncio.get_event_loop().call_later(
            delay + random.uniform(0, jitter),
            self._run_callback,
            callback,
            delay,
            jitter,
            periodic,
        )

    def _run_callback(
        self,
        callback: Callable[[], None],
        delay: float,
        jitter: float,
        periodic: float,
    ) -> None:
        try:
            callback()
        except Exception:
            logger.exception("error while executing callback")
            if periodic > 0:
                # stop the worker because the callback failed and it is periodic
                return self.bail_out("error while executing callback")
        if periodic:
            self._delayed_callback(callback, periodic, jitter, periodic)



from importlib import import_module
from typing import Any

import click


class LazyGroup(click.Group):
    def __init__(
        self,
        *,
        lazy_subcommands: dict[str, str] | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.lazy_subcommands = lazy_subcommands or {}

    def list_commands(self, ctx: click.Context) -> list[str]:
        commands = super().list_commands(ctx)
        commands.extend(self.lazy_subcommands)
        return sorted(commands)

    def get_command(self, ctx: click.Context, cmd_name: str) -> click.Command | None:
        if cmd_name in self.lazy_subcommands:
            return self._lazy_load(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _lazy_load(self, cmd_name: str) -> click.Command:
        # lazily loading a command, first get the module name and attribute name
        import_path = self.lazy_subcommands[cmd_name]
        modname, cmd_object_name = import_path.rsplit(":", 1)
        # do the import
        mod = import_module(modname)
        # get the Command object from that module
        cmd_object = getattr(mod, cmd_object_name)
        # check the result to make debugging easier
        if not isinstance(cmd_object, click.Command):
            raise ValueError(
                f"Lazy loading of {import_path} failed by returning "
                "a non-command object"
            )
        return cmd_object
